/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package rest

import (
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"os"
	"regexp"
	"strconv"
	"strings"
	"sync/atomic"
	"time"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/crt"
	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/pem"

	"github.com/sirupsen/logrus"
)

const (
	uidParam = "uid"
	verParam = "version"
)

// ErrorDetails stores the fixed strings used in the bodies of error
// responses.
type ErrorDetails struct {
	Type   string
	Title  string
	Detail string
}

var (
	pemsRegex = regexp.MustCompile("^" + pemsPfx + "([^/]+)/([^/]+)$")

	allowedRdonly = map[string]struct{}{
		http.MethodGet:  struct{}{},
		http.MethodHead: struct{}{},
	}
	allowRdonly = "GET, HEAD"

	allowedPems = map[string]struct{}{
		http.MethodGet:    struct{}{},
		http.MethodHead:   struct{}{},
		http.MethodDelete: struct{}{},
		http.MethodPost:   struct{}{},
		http.MethodPut:    struct{}{},
	}
	allowPems = "GET, HEAD, PUT, POST, DELETE"

	jsonContentType    = "application/json"
	problemContentType = "application/problem+json"
	errCtr             = uint64(0)

	errPemsPattern = ErrorDetails{
		Type:  "/errors/pems/urlPattern",
		Title: "Invalid /v1/pems/ URL path",
		Detail: "/v1/pems/ URL path does not match " +
			"/v1/pems/{namespace}/{name}",
	}
	errPemNotFound = ErrorDetails{
		Type:   "/errors/pems/notFound",
		Title:  "PEM file not found",
		Detail: "",
	}
	errPemDeleteError = ErrorDetails{
		Type:   "/errors/pems/delete/error",
		Title:  "Error deleting PEM file",
		Detail: "",
	}
	errPemSecretNotFound = ErrorDetails{
		Type:   "/errors/pems/secretNotFound",
		Title:  "Matching Secret not found",
		Detail: "",
	}
	errPemInvalidSecret = ErrorDetails{
		Type:   "/errors/pems/invalidSecret",
		Title:  "Invalid TLS Secret",
		Detail: "",
	}
	errPemPermission = ErrorDetails{
		Type:   "/errors/pems/write/permission",
		Title:  "Insufficient permission to write PEM file",
		Detail: "",
	}
	errPemWriteErr = ErrorDetails{
		Type:   "/errors/pems/write/error",
		Title:  "Error writing PEM file",
		Detail: "",
	}
	errPemsReadErr = ErrorDetails{
		Type:   "/errors/pems/read/error",
		Title:  "Error reading PEM file",
		Detail: "",
	}
	errPemsJSONMarshal = ErrorDetails{
		Type:   "/errors/pems/marshal",
		Title:  "Error marshaling JSON from PEM file",
		Detail: "",
	}
)

// Problem Details object per RFC7807
type Problem struct {
	Type     string `json:"type"`
	Title    string `json:"title"`
	Status   int    `json:"status"`
	Detail   string `json:"detail"`
	Instance string `json:"instance"`
}

// Date format for Common Log Format
const common = "02/Jan/2006:15:04:05 -0700"

func reqLog(
	log *logrus.Logger,
	req *http.Request,
	now time.Time,
	status, bytes int,
) {
	clientIP := req.RemoteAddr
	if colon := strings.LastIndex(clientIP, ":"); colon != -1 {
		clientIP = clientIP[:colon]
	}
	log.Infof("%s - - [%s] \"%s %s %s\" %d %d\n", clientIP,
		now.Format(common), req.Method, req.RequestURI, req.Proto,
		status, bytes)
}

func errLog(log *logrus.Logger, req *http.Request, err error) {
	log.Errorf("%s \"%s %s %s\": %v", req.RemoteAddr, req.Method,
		req.RequestURI, req.Proto, err)
}

func getErrInstance() string {
	n := atomic.AddUint64(&errCtr, 1)
	return "/log/errors/" + strconv.FormatUint(n, 36)
}

type healthzHndlr struct {
	log     *logrus.Logger
	version string
}

func (h *healthzHndlr) ServeHTTP(resp http.ResponseWriter, req *http.Request) {
	now := time.Now()

	resp.Header().Set("Content-Length", "0")
	resp.Header().Set("Cache-Control", "no-store")
	status := http.StatusTeapot

	if req.URL.Path != healthzPath {
		status = http.StatusNotFound
	} else if _, ok := allowedRdonly[req.Method]; !ok {
		status = http.StatusMethodNotAllowed
		resp.Header().Set("Allow", allowRdonly)
	} else {
		status = http.StatusNoContent
		resp.Header().Del("Content-Length")
		resp.Header().Set("Application-Version", h.version)
	}
	resp.WriteHeader(status)
	reqLog(h.log, req, now, status, 0)
}

type pemsHndlr struct {
	log       *logrus.Logger
	files     *pem.Files
	crtGetter *crt.Getter
}

func (h *pemsHndlr) errorResponse(
	resp http.ResponseWriter,
	req *http.Request,
	now time.Time,
	status int,
	details ErrorDetails,
	err error,
) {
	problem := Problem{
		Type:     details.Type,
		Title:    details.Title,
		Detail:   details.Detail,
		Status:   status,
		Instance: getErrInstance(),
	}
	if err != nil {
		problem.Detail = err.Error()
	}
	h.log.Errorf("%s %s: %+v", req.Method, req.RequestURI, problem)

	body, err := json.Marshal(problem)
	if err != nil {
		resp.WriteHeader(status)
		h.log.Errorf("%s %s: cannot marshal %+v: %v", req.Method,
			req.RequestURI, problem, err)
		reqLog(h.log, req, now, status, 0)
		return
	}
	resp.Header().Set("Content-Type", problemContentType)
	resp.Header().Set("Content-Length", strconv.Itoa(len(body)))
	resp.WriteHeader(status)
	n, err := resp.Write(body)
	if err != nil {
		h.log.Errorf("%s %s: cannot write body: %v", req.Method,
			req.RequestURI, err)
	}
	reqLog(h.log, req, now, status, n)
}

func (h *pemsHndlr) allPems(
	resp http.ResponseWriter,
	req *http.Request,
	now time.Time,
) {
	if _, ok := allowedRdonly[req.Method]; !ok {
		resp.Header().Set("Allow", allowRdonly)
		resp.Header().Set("Content-Length", "0")
		resp.WriteHeader(http.StatusMethodNotAllowed)
		reqLog(h.log, req, now, http.StatusMethodNotAllowed, 0)
	}

	// XXX check Accept
	// XXX I-M-S and I-N-M
	allInfo, err := h.files.GetAllFileInfo()
	if err != nil {
		h.errorResponse(resp, req, now, http.StatusInternalServerError,
			errPemsReadErr, err)
		return
	}
	jsonBytes, err := json.Marshal(allInfo)
	if err != nil {
		h.errorResponse(resp, req, now, http.StatusInternalServerError,
			errPemsJSONMarshal, err)
		return
	}
	// XXX LastModified and ETag
	resp.Header().Set("Content-Type", jsonContentType)
	resp.Header().Set("Content-Length", strconv.Itoa(len(jsonBytes)))
	n := 0
	if req.Method == http.MethodGet {
		n, err = resp.Write(jsonBytes)
		if err != nil {
			h.log.Errorf("%s %s: cannot write body: %v", req.Method,
				req.RequestURI, err)
		}
	} else {
		resp.WriteHeader(http.StatusOK)
	}
	reqLog(h.log, req, now, http.StatusOK, n)
}

func (h *pemsHndlr) pemInfo(
	ns, name, uid, version string,
	resp http.ResponseWriter,
	req *http.Request,
	now time.Time,
) {
	// XXX check Accept
	info, found, err := h.files.GetFileInfo(ns, name, uid, version)
	if !found {
		h.errorResponse(resp, req, now, http.StatusNotFound,
			errPemNotFound, err)
		return
	}
	if err != nil {
		h.errorResponse(resp, req, now, http.StatusInternalServerError,
			errPemsReadErr, err)
		return
	}
	resp.Header().Set("ETag",
		info.K8sSecret.UID+"/"+info.K8sSecret.Version)
	resp.Header().Set("Last-Modified",
		info.Meta.ModTime.UTC().Format(http.TimeFormat))

	inm := req.Header.Get("If-None-Match")
	if inm != "" {
		tags := strings.Split(strings.TrimPrefix(inm, "W/"), "/")
		if len(tags) == 2 && tags[0] == info.K8sSecret.UID &&
			tags[1] == info.K8sSecret.Version {
			resp.Header().Del("Content-Length")
			resp.WriteHeader(http.StatusNotModified)
			reqLog(h.log, req, now, http.StatusNotModified, 0)
			return
		}
	}
	ims := req.Header.Get("If-Modified-Since")
	if ims != "" {
		since, err := time.Parse(http.TimeFormat, ims)
		if err != nil {
			h.log.Errorf("%s %s: cannot parse If-Modified-Since "+
				"\"%s\": %v", req.Method, req.RequestURI, ims,
				err)
		} else if !info.Meta.ModTime.UTC().Truncate(time.Second).
			After(since.UTC()) {
			resp.Header().Del("Content-Length")
			resp.WriteHeader(http.StatusNotModified)
			reqLog(h.log, req, now, http.StatusNotModified, 0)
			return
		}
	}

	jsonBytes, err := json.Marshal(info)
	if err != nil {
		h.errorResponse(resp, req, now, http.StatusInternalServerError,
			errPemsJSONMarshal, err)
		return
	}
	resp.Header().Set("Content-Type", jsonContentType)
	resp.Header().Set("Content-Length", strconv.Itoa(len(jsonBytes)))
	n := 0
	if req.Method == http.MethodGet {
		n, err = resp.Write(jsonBytes)
		if err != nil {
			h.log.Errorf("%s %s: cannot write body: %v", req.Method,
				req.RequestURI, err)
		}
	} else {
		resp.WriteHeader(http.StatusOK)
	}
	reqLog(h.log, req, now, http.StatusOK, n)
}

func (h *pemsHndlr) ServeHTTP(resp http.ResponseWriter, req *http.Request) {
	now := time.Now()

	if req.URL.Path == pemsAll {
		h.allPems(resp, req, now)
		return
	}

	resp.Header().Set("Content-Length", "0")
	status := http.StatusTeapot

	matches := pemsRegex.FindStringSubmatch(req.URL.Path)
	if matches == nil {
		h.errorResponse(resp, req, now, http.StatusNotFound,
			errPemsPattern, nil)
		return
	}

	if _, ok := allowedPems[req.Method]; !ok {
		status = http.StatusMethodNotAllowed
		resp.Header().Set("Allow", allowPems)
		resp.WriteHeader(status)
		reqLog(h.log, req, now, status, 0)
		return
	}
	ns, name := matches[1], matches[2]
	uid := req.URL.Query().Get(uidParam)
	version := req.URL.Query().Get(verParam)

	if req.Method == http.MethodGet || req.Method == http.MethodHead {
		h.pemInfo(ns, name, uid, version, resp, req, now)
		return
	}

	if req.Method == http.MethodDelete {
		if exist, err := h.files.Delete(ns, name); !exist ||
			(err != nil && os.IsNotExist(err)) {
			h.errorResponse(resp, req, now, http.StatusNotFound,
				errPemNotFound, err)
			return
		} else if err != nil {
			h.errorResponse(resp, req, now,
				http.StatusInternalServerError,
				errPemDeleteError, err)
			return
		}
		status = http.StatusNoContent
		resp.Header().Del("Content-Length")
		resp.WriteHeader(status)
		reqLog(h.log, req, now, status, 0)
		return
	}

	// XXX If-Match: uid/version
	// XXX If-Unmodified-Since: compare file mtime
	have := h.files.Have(ns, name, uid, version)
	if have && req.Method == http.MethodPost {
		status = http.StatusConflict
		resp.WriteHeader(status)
		reqLog(h.log, req, now, status, 0)
		return
	} else if h.files.Check(ns, name, uid, version) {
		status = http.StatusNoContent
		resp.Header().Del("Content-Length")
		resp.WriteHeader(status)
		reqLog(h.log, req, now, status, 0)
		return
	}
	if found, valid, err := h.files.Write(ns, name, uid, version); !found {
		msg := fmt.Sprintf("Secret matching %s/%s", ns, name)
		if uid != "" {
			msg += " uid=" + uid
		}
		if version != "" {
			msg += " version=" + version
		}
		msg += " not found"
		if err != nil {
			msg += ": " + err.Error()
		}
		errr := errors.New(msg)
		h.errorResponse(resp, req, now, http.StatusNotFound,
			errPemSecretNotFound, errr)
		return
	} else if !valid {
		h.errorResponse(resp, req, now, http.StatusForbidden,
			errPemInvalidSecret, err)
		return
	} else if err != nil {
		if os.IsPermission(err) {
			h.errorResponse(resp, req, now,
				http.StatusInternalServerError,
				errPemPermission, err)
		} else {
			h.errorResponse(resp, req, now,
				http.StatusInternalServerError,
				errPemWriteErr, err)
		}
		return
	}
	if !have {
		status = http.StatusCreated
	} else {
		status = http.StatusNoContent
		resp.Header().Del("Content-Length")
	}
	// XXX Location: /v1/pems/ns/name?uid=uid&version=version
	// XXX ETag: uid/version
	// XXX LastModified: file mtime
	resp.WriteHeader(status)
	reqLog(h.log, req, now, status, 0)
}
